from io import BytesIO
from io import StringIO
from typing import Any
from typing import Callable

from actionpack import Action
from actionpack.action import Name
from actionpack.action import Outcome


class FakeAction(Action[Name, Outcome]):

    result = 'Performing Action.'

    @staticmethod
    def return_result():
        return FakeAction.result

    def __init__(
        self,
        name: Name = None,
        typecheck: Any = None,
        instruction_provider: Callable[[], Any] = None,
    ):
        if typecheck:
            raise TypeError(str(typecheck))

        self.name = name
        self.instruction_provider = instruction_provider
        self.state = {'this': 'state'}
        setattr(self, 'instruction', instruction_provider if instruction_provider else FakeAction.return_result)


class FakeFile:
    def __init__(self, contents=None, mode: str = None):
        if not contents:
            contents = bytes()
        self.buffer = BytesIO(contents) if isinstance(contents, bytes) else StringIO(contents)
        self.buffer.read()
        self.mode = mode

    def read(self):
        self.buffer.seek(0)
        return self.buffer.read()

    def write(self, data):
        if self.mode in ['wb', 'w']:
            self.buffer.seek(0)
            self.buffer.truncate()
        self.buffer.write(data)
        return len(data)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.closed = True



from collections import OrderedDict
from inspect import signature
from typing import Callable
from typing import Optional

from actionpack import Action
from actionpack.action import ActionType
from actionpack.actions import Call
from actionpack.utils import Closure
from actionpack.utils import first
from actionpack.utils import key_for
from actionpack.utils import swap


class Pipeline(Action):

    def __init__(self, action: Action, *action_types: ActionType, should_raise: bool = False):
        self.action = action
        self.should_raise = should_raise

        for action_type in action_types:
            if not isinstance(action_type, ActionType):
                raise TypeError(f'Must be an {ActionType.__name__}: {action_type}')

        self.action_types = action_types
        self._action_types = iter(action_types)

    def instruction(self):
        return self.flush(self.action).perform(should_raise=self.should_raise).value

    # recursive
    def flush(self, given_action: Optional[Action] = None) -> Action:
        next_action_type = next(self)

        if next_action_type:
            params_dict = OrderedDict(signature(next_action_type.__init__).parameters.items())
            params_dict.pop('self', None)
            params = list(params_dict.keys())
            keyed_result = dict(zip(params, [given_action.perform(should_raise=self.should_raise).value]))
            if next_action_type.__name__ == Pipeline.Fitting.__name__:
                params_dict = OrderedDict(signature(next_action_type.action.__init__).parameters.items())
                params_dict.pop('self', None)
                params = list(params_dict.keys())
                conduit = first(params)
                key = key_for(Pipeline.Receiver, dict(next_action_type.kwargs))
                if conduit in next_action_type.kwargs and key:
                    params = swap(params, 0, params.index(key))
                keyed_result = dict(zip(params, [keyed_result['action']]))
                if conduit == 'closure':
                    keyed_result['closure'] = Closure(next_action_type.enclose, keyed_result['closure'])
                next_action_type.kwargs.update(keyed_result)
                next_action = next_action_type.action(**next_action_type.kwargs)
                next_action_type.kwargs[key] = Pipeline.Receiver
            else:
                next_action = next_action_type(**keyed_result)

            return self.flush(next_action)

        return given_action

    def __iter__(self):
        return self._action_types

    def __next__(self):
        try:
            return next(self._action_types)
        except StopIteration:
            self._action_types = iter(self.action_types)

    class Receiver:
        pass

    class Fitting(type):

        @staticmethod
        def init(
            action: ActionType,
            should_raise: bool = False,
            enclose: Callable = None,
            reaction: Call = None,
            *args, **kwargs
        ):
            pass

        @staticmethod
        def instruction(instance):
            action_performance = instance.action.perform(should_raise=instance.should_raise)
            if action_performance.successful and instance.reaction:
                return instance.reaction.perform(should_raise=instance.should_raise)
            return action_performance

        def __new__(
            mcs,
            action: ActionType,
            should_raise: bool = False,
            enclose: Callable = None,
            reaction: Call = None,
            **kwargs
        ):
            dct = dict()
            dct['__init__'] = Pipeline.Fitting.init
            dct['instruction'] = Pipeline.Fitting.instruction
            dct['action'] = action
            dct['enclose'] = enclose
            dct['should_raise'] = should_raise
            dct['reaction'] = reaction
            dct['kwargs'] = kwargs
            cls = type(Pipeline.Fitting.__name__, (Action,), dct)
            return cls



import pickle

from unittest import TestCase

from actionpack.action import Result
from actionpack.actions import Call
from actionpack.utils import Closure
from actionpack.utils import pickleable


class CallTest(TestCase):

    @classmethod
    def function(*args, **kwargs) -> tuple:
        return args[1:], kwargs

    def setUp(self):
        self.arg = 'arg'
        self.kwarg = 'kwarg'
        self.closure = Closure(self.function, self.arg, kwarg=self.kwarg)
        self.action = Call(closure=self.closure)

    def test_can_Call(self):
        result = self.action.perform()
        self.assertIsInstance(result, Result)
        self.assertEqual(result.value, ((self.arg,), {self.kwarg: self.kwarg}))

    def test_Call_instantiation_accepts_only_Closure_objects(self):
        result = Call('not a closure').perform()
        self.assertFalse(result.successful)
        self.assertIsInstance(result.value, TypeError)

    def test_Call_can_handle_Closures_with_iterator_args(self):
        iterable = [1, 1, 1]
        closure = Closure(sum, iter(iterable))
        call = Call(closure)
        result = call.perform()
        new_result = call.perform()
        self.assertTrue(result.successful)
        self.assertEqual(sum(iterable), result.value)
        self.assertEqual(0, new_result.value)  # since iterable exhausted

    def test_serializes_as_expected(self):
        self.assertEqual(repr(self.action), '<Call(function)>')

    def test_can_pickle(self):
        pickled = pickleable(self.action)
        unpickled = pickle.loads(pickled)

        self.assertTrue(pickleable(self.action))
        self.assertEqual(unpickled.__dict__, self.action.__dict__)



from __future__ import annotations
from functools import partialmethod
from oslash import Left
from oslash import Right
from oslash.either import Either
from string import Template
from subprocess import PIPE
from subprocess import run
from sys import executable as python
from threading import RLock
from typing import Callable
from typing import Generic
from typing import Optional
from typing import TypeVar
from typing import Union

from actionpack.utils import microsecond_timestamp
from actionpack.utils import synchronized


Outcome = TypeVar('Outcome')
Name = TypeVar('Name')
ResultValue = Union[Outcome, Exception]

T = TypeVar('T')
V = TypeVar('V')


class Result(Generic[Outcome]):

    _immutables = ('successful', 'produced_at', 'value',)

    def __init__(
        self,
        outcome: Either,
        timestamp_provider: Callable[[], int] = microsecond_timestamp
    ):
        self.value: Optional[ResultValue[Outcome]] = None
        if type(outcome) in [Left, Right]:
            self.value = outcome.value
            self.successful = True if isinstance(outcome, Right) else False
        else:
            raise self.OutcomeMustBeOfTypeEither

        self.produced_at = timestamp_provider()

    def __repr__(self):
        outcome = 'success' if self.successful else 'failure'
        return f'<Result|{outcome}[{type(self.value).__name__}]>'

    def __setattr__(self, name, value):
        if name in self._immutables and getattr(self, name, None) is not None:
            raise AttributeError(f'Cannot modify `{name}`')
        else:
            return super(Result, self).__setattr__(name, value)

    def __delattr__(self, name):
        if name in self._immutables:
            raise AttributeError(f'Cannot modify `{name}`')
        else:
            return super(Result, self).__delattr__(name)

    class OutcomeMustBeOfTypeEither(Exception):
        pass


class ActionType(type):

    def __call__(self, *args, reaction: Optional[Action[T, V]] = None, **kwargs):
        failure = None

        def instruction():
            raise failure

        try:
            instance = super().__call__(*args, **kwargs)
            instance.__reaction = reaction
        except Exception as e:
            return Action.Construct(e)

        params = args + tuple(kwargs.values())
        for param in params:
            if issubclass(type(param), Exception):
                failure = param
                setattr(instance, 'instruction', instruction)

        return instance


def partialaction(name, parent: T, **kwargs) -> T:
    partial__init__ = partialmethod(parent.__init__, **kwargs)
    return type(name, (parent,), {'__init__': partial__init__})


class Action(Generic[Name, Outcome], metaclass=ActionType):

    _name: Optional[Name] = None

    lock = RLock()
    requirements = tuple()

    def instruction(self):
        pass

    @synchronized(lock)
    def perform(
        self,
        should_raise: bool = False,
        timestamp_provider: Callable[[], int] = microsecond_timestamp
    ) -> Result[Outcome]:
        return self._perform(should_raise, timestamp_provider)

    def validate(self):
        return self

    def set(self, **kwargs) -> Action[Name, Outcome]:
        self._name: Optional[Name] = kwargs.get('name')
        return self

    @property
    def name(self) -> Optional[Name]:
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @name.deleter
    def name(self):
        del self._name

    def check_dependencies(self, cls):
        for requirement in cls.requirements:
            if not hasattr(cls, requirement):
                Action.DependencyCheck(cls, requirement)

    def _perform(
        self,
        should_raise: bool = False,
        timestamp_provider: Callable[[], int] = microsecond_timestamp
    ) -> Result[Outcome]:
        if not callable(self.instruction):
            outcome = Left(TypeError(f'Must be callable: {self.instruction}'))
        else:
            try:
                outcome = Right(self.validate().instruction())
                if issubclass(type(outcome.value), Exception):
                    raise outcome.value
            except Exception as e:
                if should_raise:
                    raise e
                outcome = Left(e)
            finally:
                if self._ActionType__reaction:
                    self._ActionType__reaction.perform(should_raise=should_raise)

        return Result(outcome, timestamp_provider)

    def __init_subclass__(cls, requires=None):
        if requires:
            cls.requirements += requires

    def __getstate__(self):
        return vars(self)

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            raise self.__class__.NotComparable(f'{str(self)} ≠ {str(other)}')

        return self.__getstate__() == other.__getstate__()

    def __repr__(self):
        tmpl = Template(f'<{self.__class__.__name__}$name>')
        return tmpl.substitute(name=f'|name="{self.name}"') if self.name else tmpl.substitute(name='')

    class NotComparable(Exception):
        pass

    class Invalid(Exception):
        pass

    class Construct:
        def __init__(self, failure: Exception):
            self.failure = failure

        def perform(self, should_raise: bool = False) -> Result:
            return Result(Left(self.failure))

        def __repr__(self):
            return f'<Action.Construct[{self.failure.__class__.__name__}]>'

    class DependencyCheck:
        def __init__(self, cls, requirement: str = None):
            if not requirement:
                raise self.WhichPackage('do you want to check? Please specify a requirment kwarg.')

            result = run([python, '-m', 'pip', 'show', requirement], stdout=PIPE, stderr=PIPE)
            if result.returncode != 0:
                raise self.PackageMissing(f'so please install "{requirement}" to proceed.')

            setattr(cls, requirement, __import__(requirement))

        class PackageMissing(Exception):
            pass

        class WhichPackage(Exception):
            pass



from __future__ import annotations
from pathlib import Path
from typing import Union

from actionpack import Action
from actionpack.action import Name


FileContents = Union[bytes, str]


class Remove(Action[Name, FileContents]):

    def __init__(self, filename: str, tail: bool = False):
        self.path = Path(filename)
        self.tail = tail

    def instruction(self) -> FileContents:
        with open(str(self.path), 'r+') as file:
            lines = file.readlines()
            if lines:
                to_remove = lines.pop(-1) if self.tail else lines.pop(0)
                file.seek(0)
                for line in lines:
                    file.write(line)
                file.truncate()
        return to_remove if lines else ''

    def validate(self) -> Remove[Name, FileContents]:
        if not self.path.exists():
            raise FileNotFoundError(str(self.path))
        if self.path.is_dir():
            raise IsADirectoryError(str(self.path))
        return self



import pickle

from datetime import datetime
from unittest import TestCase
from unittest.mock import ANY
from unittest.mock import patch

from actionpack import Action
from actionpack.action import Result
from actionpack.actions import MakeRequest
from actionpack.actions import RetryPolicy
from actionpack.utils import pickleable
from tests.actionpack.actions import FakeResponse


class RetryPolicyTest(TestCase):

    def setUp(self):
        self.max_retries = 2
        self.action = RetryPolicy[str, str](
            MakeRequest('GET', 'http://localhost'),
            max_retries=self.max_retries
        )

    @patch('requests.Session.send')
    def test_RetryPolicy_can_succeed(self, mock_session_send):
        result = self.action.perform(should_raise=True)

        self.assertIsInstance(result, Result)
        self.assertTrue(result.successful)

    @patch('requests.Session.send')
    def test_can_enact_RetryPolicy(self, mock_session_send):
        mock_session_send.side_effect = Exception('something went wrong :/')
        result = self.action.perform()

        self.assertEqual(self.action.retries, self.max_retries)
        self.assertIsInstance(result, Result)
        self.assertIsInstance(result.value, RetryPolicy.Expired)

    @patch('requests.Session.send')
    def test_can_handle_inappropriately_enacted_RetryPolicy(self, mock_session_send):
        mock_session_send.side_effect = Exception('something went wrong :/')
        with self.assertRaises(RetryPolicy.Invalid):
            self.action.enact(counter=-10)

    @patch('requests.Session.send')
    def test_can_validate_expired_RetryPolicy(self, mock_session_send):
        exceptions = [Exception('something went wrong :/')] * self.max_retries
        mock_session_send.side_effect = exceptions
        result = self.action.perform()

        self.assertIsInstance(result.value, RetryPolicy.Expired)
        with self.assertRaises(RetryPolicy.Expired):
            self.action.validate()

    @patch('requests.Session.send')
    @patch('actionpack.actions.RetryPolicy.enacted')
    def test_can_validate_enacted_RetryPolicy(self, mock_validation, mock_session_send):
        exceptions = [Exception('something went wrong :/'), ANY]
        mock_session_send.side_effect = exceptions
        mock_validation.side_effect = [ANY, RetryPolicy.Enacted('already started.')]
        result = self.action.perform()

        self.assertIsInstance(result.value, RetryPolicy.Enacted)
        with self.assertRaises(RetryPolicy.Enacted):
            self.action.validate()

    @patch('requests.Session.send')
    def test_can_enact_RetryPolicy_that_ultimately_succeeds(self, mock_session_send):
        mock_session_send.side_effect = [
            Exception('something went wrong :/'),
            FakeResponse('something went right :)'.encode())
        ]
        result = self.action.perform()

        self.assertIsInstance(result, Result)
        self.assertEqual(self.action.retries, 1)

    def test_no_retries_before_RetryPolicy_enacted(self):
        with self.assertRaises(AttributeError):
            self.action.entries

    def test_validate(self):
        invalid_num_retries = -1
        self.action._retries = invalid_num_retries
        validated_action = self.action.validate()

        self.assertEqual(validated_action.retries, invalid_num_retries)

    @patch('requests.Session.send')
    def test_can_record_retry_history_if_not_successful(self, mock_session_send):
        exceptions = [Exception('something went wrong :/')] * (self.max_retries + 1)
        mock_session_send.side_effect = exceptions
        action = RetryPolicy[str, str](
            action=MakeRequest('GET', 'http://localhost'),
            max_retries=self.max_retries,
            should_record=True
        )
        action.perform()

        for attempt in action.attempts:
            self.assertFalse(attempt.successful)
        self.assertListEqual([attempt.value for attempt in action.attempts], exceptions)

    @patch('requests.Session.send')
    def test_can_record_retry_history_if_successful(self, mock_session_send):
        results = ['SUCCEEDED!']
        mock_session_send.side_effect = results
        action = RetryPolicy[str, str](
            action=MakeRequest('GET', 'http://localhost'),
            max_retries=self.max_retries,
            should_record=True
        )
        action.perform()

        for attempt in action.attempts:
            self.assertTrue(attempt.successful)
        self.assertListEqual([attempt.value for attempt in action.attempts], results)

    @patch('requests.Session.send')
    def test_can_enact_with_linear_backoff(self, mock_session_send):
        delay = 0.2  # seconds
        delay_microseconds = delay * 1E6
        mock_session_send.side_effect = [Exception('failed.')] * (self.max_retries + 1)
        action = RetryPolicy[str, str](
            action=MakeRequest('GET', 'http://localhost'),
            max_retries=4,
            delay_between_attempts=delay,
            backoff='LINEAR',
            should_record=True,
        )
        result = action.perform(timestamp_provider=lambda: round(datetime.now().timestamp()))
        times = [
            attempt.produced_at - action.attempts[i - 1].produced_at
            for i, attempt in reversed(list(enumerate(action.attempts))) if i != 0
        ]

        self.assertFalse(result.successful)
        expected_target_factors = [4.0, 3.0, 2.0, 1.0]  # linear sequence
        actual_target_factors = []
        for time in times:
            fraction = time / delay_microseconds
            target_factor = round(fraction) * delay_microseconds
            actual_target_factors.append(target_factor // delay_microseconds)
            actual_jitter_percentage = time / (target_factor * delay_microseconds)
            self.assertLessEqual(actual_jitter_percentage, action.jitter_percentage)
        self.assertListEqual(actual_target_factors, expected_target_factors)

    @patch('requests.Session.send')
    def test_can_enact_with_exponential_backoff(self, mock_session_send):
        delay = 0.2  # seconds
        delay_microseconds = delay * 1E6
        mock_session_send.side_effect = [Exception('failed.')] * (self.max_retries + 1)
        action = RetryPolicy[str, str](
            action=MakeRequest('GET', 'http://localhost'),
            max_retries=4,
            delay_between_attempts=delay,
            backoff='EXPONENTIAL',
            should_record=True,
        )
        result = action.perform(timestamp_provider=lambda: round(datetime.now().timestamp()))
        times = [
            attempt.produced_at - action.attempts[i - 1].produced_at
            for i, attempt in reversed(list(enumerate(action.attempts))) if i != 0
        ]

        self.assertFalse(result.successful)
        expected_target_factors = [8.0, 4.0, 2.0, 1.0]  # exponential sequence
        actual_target_factors = []
        for time in times:
            fraction = time / delay_microseconds
            target_factor = round(fraction) * delay_microseconds
            actual_target_factors.append(target_factor // delay_microseconds)
            actual_jitter_percentage = time / (target_factor * delay_microseconds)
            self.assertLessEqual(actual_jitter_percentage, action.jitter_percentage)
        self.assertListEqual(actual_target_factors, expected_target_factors)

    @patch('requests.Session.send')
    def test_delay_is_bypassed_after_expiration(self, mock_session_send):
        results = ['SUCCEEDED!']
        delay = 300  # seconds
        mock_session_send.side_effect = results
        action = RetryPolicy[str, str](
            action=MakeRequest('GET', 'http://localhost'),
            max_retries=0,
            delay_between_attempts=delay,
            should_record=True,
        )
        timestamp_provider = lambda: round(datetime.now().timestamp())
        result = action.perform(timestamp_provider=timestamp_provider)
        assert result.produced_at < timestamp_provider() + delay

    def test_instantiation_fails_given_invalid_max_retries(self):
        action = RetryPolicy[str, str](action=MakeRequest('GET', 'http://localhost'), max_retries=-1)
        self.assertIsInstance(action, Action.Construct)
        result = action.perform()
        self.assertFalse(result.successful)
        self.assertIsInstance(result.value, RetryPolicy.Invalid)

    @patch('requests.Session.send')
    def test_can_serialize(self, mock_session_send):
        action = RetryPolicy[str, str](
            action=MakeRequest('GET', 'http://localhost'),
            max_retries=0,
        )
        self.assertEqual(repr(action), '<RetryPolicy(1 x <MakeRequest>)>')
        self.assertEqual(repr(self.action), '<RetryPolicy(3 x <MakeRequest>)>')

    @patch('requests.Session.send')
    def test_can_pickle(self, mock_session_send):
        mock_session_send.side_effect = NotImplemented
        self.assertTrue(pickleable(self.action))

        pickled = pickleable(self.action)
        unpickled = pickle.loads(pickled)

        self.assertTrue(pickleable(self.action))
        self.assertEqual(unpickled.__dict__, self.action.__dict__)



from __future__ import annotations

import math
import random

from enum import Enum
from string import Template
from time import sleep

from actionpack import Action
from actionpack.action import Result
from actionpack.action import Name
from actionpack.action import Outcome
from actionpack.utils import tally


class RetryPolicy(Action[Name, Outcome]):
    def __init__(
        self,
        action: Action[Name, Outcome],
        max_retries: int,
        delay_between_attempts: int = 0,
        max_delay: int = 3600,
        backoff: str = 'CONSTANT',
        jitter_percentage: float = 1.0,
        should_record: bool = False
    ):
        if not isinstance(max_retries, int) or max_retries < 0:
            raise self.Invalid(f'The number of max_retries must be greater than zero. Given max_retries={max_retries}.')
        if jitter_percentage not in range(101):
            raise self.Invalid(f'The number of max_retries must be greater than zero. Given max_retries={max_retries}.')

        self.action = action
        self.backoff = RetryPolicy.Backoff(backoff)
        self.delay_between_attempts = delay_between_attempts
        self.jitter_percentage = jitter_percentage
        self.max_delay = max_delay
        self.max_retries = max_retries
        self.should_record = should_record

        if self.should_record:
            self.attempts: list[Result[Outcome]] = []

    def instruction(self) -> Outcome:
        return self.enact()

    def validate(self) -> RetryPolicy:
        if self.expired:
            raise RetryPolicy.Expired(f'Gave up on {str(self.action)} after {self.max_retries + 1} attempts.')
        if self.enacted:
            raise RetryPolicy.Enacted(f'{str(self.action)} already attempted. Will not perform.')
        return self

    @property
    def retries(self) -> int:
        try:
            return self._retries
        except AttributeError:
            pass
        return -1

    @property
    def enacted(self):
        return self.retries >= 0

    @property
    def expired(self):
        return self.retries >= self.max_retries

    def enact(self, counter: int = -1) -> Outcome:
        if not isinstance(counter, int) or counter < -1:
            raise self.Invalid(f'Cannot proceed with given `counter` param value: {counter}.')

        for _tally in tally(1 + self.max_retries):
            if self.expired:
                break  # pragma: no cover
            attempt = self.action.perform()
            counter += _tally
            with_delay = self.backoff.calculate(
                counter, self.delay_between_attempts, self.jitter_percentage
            )
            self._retries = counter
            if self.should_record:
                self.attempts.append(attempt)
            if attempt.successful:
                return attempt.value
            sleep(with_delay)

        raise RetryPolicy.Expired(f'Max retries exceeded: {self.max_retries}.')

    def __repr__(self):
        tmpl = Template('<$class_name($total_attempts x $action_name$delay)>')
        return tmpl.substitute(
            class_name=self.__class__.__name__,
            total_attempts=self.max_retries + 1,
            action_name=str(self.action),
            delay='' if not self.delay_between_attempts else f' | {self.delay_between_attempts}s delay'
        )

    class Expired(Exception):
        pass

    class Enacted(Exception):
        pass

    class Backoff(Enum):
        CONSTANT = 'CONSTANT'
        LINEAR = 'LINEAR'
        EXPONENTIAL = 'EXPONENTIAL'

        def calculate(self, coefficient: float, base: int, jitter_percentage: float):
            if base == 0:
                return 0
            factor, is_fraction = self._rectify(base)
            base = abs(factor * base if is_fraction else base)
            coefficient = abs(coefficient)

            approaches = {
                self.CONSTANT: (coefficient ** 0) * base,       # 2, 2, 2, 2, ...
                self.LINEAR: ((coefficient + 1) ** 1) * base,   # 2, 4, 6, 8, ...
                self.EXPONENTIAL: (base ** coefficient) * base  # 2, 4, 8, 16, ...
            }
            delay = approaches[self] / factor if is_fraction else approaches[self]
            jitter_fraction = (jitter_percentage / 100) * base
            jitter = random.uniform(-jitter_fraction, jitter_fraction)
            return delay + jitter

        def _rectify(self, base: float):
            order_of_magnitude = math.floor(math.log(abs(base), 10))
            is_fraction = order_of_magnitude < 1
            if is_fraction:
                factor = 1 * 10 ** (abs(order_of_magnitude))
            else:
                factor = 1
            return factor, is_fraction



import pickle

from types import ModuleType
from unittest import TestCase
from unittest.mock import patch
from validators import ValidationFailure

from actionpack.action import Result
from actionpack.actions import MakeRequest
from actionpack.utils import pickleable
from tests.actionpack.actions import FakeResponse


class MakeRequestTest(TestCase):

    def setUp(self):
        self.action = MakeRequest('GET', 'http://localhost')

    @patch('requests.Session.send')
    def test_can_MakeRequest(self, mock_session_send):
        mock_session_send.return_value = FakeResponse()
        self.assertIsInstance(self.action.perform(), Result)

    @patch('requests.Session.send')
    def test_can_MakeRequest_even_if_failure(self, mock_session_send):
        mock_session_send.return_value = FakeResponse(status=500)

        first_action = MakeRequest('SUP', 'http://localhost')
        first_result = first_action.perform()
        self.assertIsInstance(first_result, Result)
        self.assertIsInstance(first_result.value, ValueError)

        second_action = MakeRequest('GET', 'http://some/invalid/url')
        second_result = second_action.perform()
        self.assertIsInstance(second_result, Result)
        self.assertIsInstance(second_result.value, ValidationFailure)

    @patch('requests.Session.send')
    def test_MakeRequest_has_portable_dependency(self, mock_session_send):
        mock_session_send.return_value = FakeResponse(status=500)

        self.action.perform()
        self.assertIsInstance(self.action.requests, ModuleType)

    def test_can_pickle(self):
        pickled = pickleable(self.action)
        unpickled = pickle.loads(pickled)

        self.assertTrue(pickleable(self.action))
        self.assertEqual(unpickled.__dict__, self.action.__dict__)



import pickle

from functools import reduce
from oslash import Left
from oslash import Right
from threading import Thread
from unittest import TestCase
from unittest.mock import patch

from actionpack import Action
from actionpack import partialaction
from actionpack.action import Result
from actionpack.utils import pickleable
from tests.actionpack import FakeAction
from tests.actionpack import FakeFile
from tests.actionpack.actions import FakeWrite


class ActionTest(TestCase):

    def raise_failure(self):
        raise self.exception

    def setUp(self) -> None:
        self.exception = Exception('something went wrong :/')

    def test_Action_returning_exception_does_not_have_successful_Result(self):
        result = FakeAction(instruction_provider=lambda: self.exception).perform()
        self.assertIsInstance(result, Result)
        self.assertFalse(result.successful)

    def test_Action_initialized_with_Exception_has_unsuccessful_Result(self):
        result = FakeAction(self.exception).perform()
        self.assertIsInstance(result, Result)
        self.assertFalse(result.successful)

    def test_Action_produces_Result_result_when_performed(self):
        result = FakeAction().perform()
        self.assertIsInstance(result, Result)
        self.assertEqual(result.value, FakeAction.result)

    def test_Action_produces_Result_if_exception_raised_when_performed(self):
        result = FakeAction(instruction_provider=self.raise_failure).perform()
        self.assertIsInstance(result, Result)
        self.assertEqual(result.value, self.exception)

    def test_Action_can_raise_exception(self):
        with self.assertRaises(type(self.exception)):
            FakeAction(instruction_provider=self.raise_failure).perform(should_raise=True)

    def test_can_create_partial_Action(self):
        def instruction_provider():
            return some_result_value

        some_result_value = 'success!'
        PartialAction = partialaction(
            'FakeActionWithInstruction',
            FakeAction,
            instruction_provider=instruction_provider
        )
        self.assertEqual(PartialAction().perform().value, some_result_value)

    def test_can_determine_if_Result_was_successful(self):
        success = FakeAction().perform()
        failure = FakeAction(instruction_provider=self.raise_failure).perform()

        self.assertTrue(success.successful)
        self.assertFalse(failure.successful)

    def test_can_react_to_failure(self):
        vessel = []
        contents = 'contents'

        def fill():
            vessel.append(contents)

        self.assertNotIn(contents, vessel)

        reaction = FakeAction(instruction_provider=fill)
        action = FakeAction(
            instruction_provider=self.raise_failure,
            reaction=reaction
        )
        result = action.perform()

        self.assertFalse(result.successful)
        self.assertIn(contents, vessel)

    @patch('oslash.Left.__init__')
    def test_can_react_to_failure_during_catastrophe(self, mock_wrapper):
        def raise_another_failure(e):
            raise e

        mock_wrapper.side_effect = raise_another_failure

        vessel = []
        contents = 'contents'

        def fill():
            vessel.append(contents)

        reaction = FakeAction(instruction_provider=fill)
        action = FakeAction(
            instruction_provider=self.raise_failure,
            reaction=reaction
        )

        with self.assertRaises(Exception):
            result = action.perform()
            self.assertFalse(result.successful)

        self.assertIn(contents, vessel)

    def test_Action_Construct(self):
        construct = FakeAction(typecheck='Action instantiation fails.')
        result = construct.perform()
        self.assertEqual(construct.failure, result.value)
        self.assertIsInstance(construct, Action.Construct)
        self.assertIsInstance(result, Result)
        self.assertIsInstance(result.value, Exception)

    def test_Action_can_be_serialized(self):
        action = FakeAction()
        pickled = pickle.dumps(action)
        unpickled = pickle.loads(pickled)

        self.assertEqual(pickleable(action), pickled)
        self.assertEqual(unpickled.result, action.result)
        self.assertEqual(unpickled.state, action.state)

    def test_can_safely_perform_Actions_concurrently(self):
        file = FakeFile(b'Hello.', 'wb')

        def perform_and_collect(action: Action, results: list):
            action.perform()
            results.append(file.read())

        initial_file_contents = file.read()
        action1 = FakeWrite(file, b' How are you?', delay=0.2)
        action2 = FakeWrite(file, b' I hope you\'re well.', delay=0.1)
        results = [initial_file_contents]
        thread1 = Thread(target=perform_and_collect, args=(action1, results))
        thread2 = Thread(target=perform_and_collect, args=(action2, results))
        threads = thread1, thread2
        [thread.start() for thread in threads]
        [thread.join() for thread in threads]

        self.assertEqual(
            reduce(lambda a, b: a + b, results),
            initial_file_contents + action1.to_write + action2.to_write
        )

    def test_Action_can_be_renamed(self):
        action = FakeAction()
        self.assertIsNone(action.name)
        name1 = 'new'
        action.set(name=name1)
        self.assertEqual(action.name, name1)
        name2 = 'different name'
        action.name = name2
        self.assertEqual(action.name, name2)

    def test_Action_has_unimplemented_instruction(self):
        Action().instruction()

    def test_can_delete_Action_name(self):
        action = FakeAction()
        action_name = 'gone soon.'
        action.set(name=action_name)
        self.assertEqual(action.name, action_name)
        del action.name

    def test_Action_perform_fails_for_non_callable_instruction(self):
        action = FakeAction(instruction_provider='nope.')
        result = action.perform()
        self.assertFalse(result.successful)
        self.assertIsInstance(result.value, TypeError)

    def test_can_compare_Actions(self):
        fake_action = FakeAction()
        self.assertTrue(fake_action == FakeAction())
        with self.assertRaises(Action.NotComparable):
            fake_action == '<fake_action>'

    def test_Action_Construct_has_string_representation(self):
        failed_action_instantiation = FakeAction(typecheck=True)
        result = failed_action_instantiation.perform()
        self.assertIsInstance(failed_action_instantiation, Action.Construct)
        self.assertEqual(repr(failed_action_instantiation), f'<Action.Construct[{result.value.__class__.__name__}]>')

    def test_partial_Action_has_string_representation(self):
        partial_action_name = 'FakeActionWithInstruction'
        PartialAction = partialaction(
            partial_action_name,
            FakeAction,
            instruction_provider=lambda: 'success!'
        )
        self.assertEqual(repr(PartialAction()), f'<{partial_action_name}>')

    def test_DependencyCheck_fails_if_package_missing(self):
        FakeAction.requirements = ('not-a-real-packing-never-will-be',)
        with self.assertRaises(Action.DependencyCheck.PackageMissing):
            FakeAction().check_dependencies(FakeAction)

    def test_DependencyCheck_fails_if_requirement_absent(self):
        with self.assertRaises(Action.DependencyCheck.WhichPackage):
            Action.DependencyCheck(FakeAction)


class ResultTest(TestCase):

    def raise_failure(self):
        raise self.exception

    def test_Result_success_is_immutable(self):
        success = FakeAction().perform()
        failure = FakeAction(instruction_provider=self.raise_failure).perform()

        with self.assertRaises(AttributeError):
            success.successful = 'nah.'

        with self.assertRaises(AttributeError):
            failure.successful = 'maybe?'

    def test_Result_has_timestamp(self):
        result = FakeAction(instruction_provider=lambda: 'succeeded').perform(
            timestamp_provider=lambda: 0
        )

        self.assertTrue(result.successful)
        self.assertEqual(result.produced_at, 0)

    def test_cannot_instantiate_without_Either(self):
        with self.assertRaises(Result.OutcomeMustBeOfTypeEither):
            Result('not an Either type')

    def test_cannot_change_immutable_attributes(self):
        result = Result(Right('correct.'))
        with self.assertRaises(AttributeError):
            result.produced_at = 123
        with self.assertRaises(AttributeError):
            result.successful = False
        with self.assertRaises(AttributeError):
            result.value = 'some other value'

        result.who_cares = 'right?'  # should not raise since not declared immutable

    def test_cannot_delete_immutable_attributes(self):
        result = Result(Right('correct.'))
        with self.assertRaises(AttributeError):
            del result.produced_at
        with self.assertRaises(AttributeError):
            del result.successful
        with self.assertRaises(AttributeError):
            del result.value

        result.who_cares = 'right?'
        del result.who_cares  # should not raise since not declared immutable

    def test_can_serialize_result(self):
        successful_outcome = 'correct.'
        successful_result = Result(Right(successful_outcome))
        self.assertEqual(str(successful_result), f'<Result|success[{type(successful_outcome).__name__}]>')

        unsuccessful_outcome = RuntimeError('incorrect.')
        unsuccessful_result = Result(Right(unsuccessful_outcome))
        self.assertEqual(str(unsuccessful_result), f'<Result|success[{type(unsuccessful_outcome).__name__}]>')

        confused_result = Result(Left(successful_outcome))
        self.assertFalse(confused_result.successful)



from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from functools import reduce
from itertools import islice
from itertools import tee
from typing import Generic
from typing import Iterator
from typing import Iterable
from typing import Tuple

from actionpack.action import Name
from actionpack.action import Outcome
from actionpack.action import Result
from actionpack import Action


class Procedure(Generic[Name, Outcome]):

    def __init__(self, actions: Iterable[Action[Name, Outcome]]):
        if not (isinstance(actions, Iterator) or isinstance(actions, Iterable)):
            raise TypeError(f'Actions must be iterable. Received {type(actions).__name__}.')

        self.actions, self._actions, self.__actions = tee(actions, 3)

    def validate(self):
        actions, spare = tee(self.__actions, 2)
        self.__actions = spare
        for action in actions:
            if not isinstance(action, Action):
                msg = f'Procedures can only execute Actions: {str(action)}'
                raise Procedure.NotAnAction(msg)
        return self

    def execute(
        self,
        max_workers: int = 5,
        should_raise: bool = False,
        synchronously: bool = True
    ) -> Iterator[Result[Outcome]]:
        actions, spare = tee(self._actions, 2)
        self._actions = spare
        if synchronously:
            for action in actions:
                yield action.perform(should_raise=should_raise) if should_raise else action.perform()
        else:
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                futures = {executor.submit(action._perform, should_raise=should_raise): str(action) for action in actions}
                for future in as_completed(futures):
                    yield future.result()
        return True

    def __repr__(self):
        actions, spare = tee(self.actions, 2)
        self.actions = spare
        limit = 5
        some_actions = list(islice(actions, limit))
        try:
            next(actions)
            more_actions = True
        except StopIteration:
            more_actions = False
        header = '\nProcedure for performing the following Actions:\n'
        tab = '  '
        bullet = f'{tab}* '
        action_list = reduce(lambda a, b: str(a) + f'\n{bullet}' + str(b), some_actions)
        footer = f'\n{tab}...\n' if more_actions else ''
        return header + bullet + str(action_list) + footer

    def __iter__(self):
        return self._actions

    def __next__(self):
        try:
            return next(self._actions)
        except StopIteration:
            return None

    class NotAnAction(Exception):
        pass


class KeyedProcedure(Procedure[Name, Outcome]):

    def __init__(self, actions: Iterable[Action[Name, Outcome]]):
        if not (isinstance(actions, Iterator) or isinstance(actions, Iterable)):
            raise TypeError(f'Actions must be iterable. Received {type(actions)}.')

        self.actions, self._actions, self.__actions = tee(actions, 3)

    def validate(self):
        actions, spare = tee(self.__actions, 2)
        self.__actions = spare
        for action in actions:
            if not isinstance(action, Action):
                msg = f'Procedures can only execute Actions: {str(action)}'
                raise Procedure.NotAnAction(msg)
            if action.name is None:
                msg = f'All {self.__class__.__name__} Actions must have a name: {str(action)}'
                raise KeyedProcedure.UnnamedAction(msg)
        return self

    def execute(
        self,
        max_workers: int = 5,
        should_raise: bool = False,
        synchronously: bool = True
    ) -> Iterator[Tuple[Name, Result[Outcome]]]:
        if synchronously:
            for action in self:
                yield (action.name, action.perform(should_raise=should_raise)) \
                      if should_raise else (action.name, action.perform())
        else:
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                futures = {executor.submit(action._perform, should_raise=should_raise): action for action in self}
                for future in as_completed(futures):
                    yield (futures[future].name, future.result())

    class UnnamedAction(Exception):
        pass



from actionpack.action import Action
from actionpack.procedure import Procedure
from actionpack.procedure import KeyedProcedure
from actionpack.action import partialaction


__all__ = [
    'Action',
    'Procedure',
    'KeyedProcedure',
    'partialaction'
]



from __future__ import annotations
from pathlib import Path

from actionpack import Action
from actionpack.action import Name


class Read(Action[Name, bytes]):
    def __init__(self, filename: str, output_type: type = str):
        if output_type not in [bytes, str]:
            raise TypeError(f'Must be of type bytes or str: {output_type}')
        self.output_type = output_type
        self.path = Path(filename)

    def instruction(self) -> bytes:
        return self.path.read_bytes() if self.output_type is bytes else self.path.read_text()

    def validate(self) -> Read[Name, bytes]:
        if not self.path.exists():
            raise FileNotFoundError(str(self.path))
        if self.path.is_dir():
            raise IsADirectoryError(str(self.path))
        return self



from contextlib import contextmanager
from io import IOBase as Buffer
from requests import Response
from time import sleep

from actionpack import Action
from actionpack.action import Name
from tests.actionpack import FakeFile


@contextmanager
def uncloseable(buffer: Buffer):
    """
    Context manager which turns the fd's close operation to no-op for the duration of the context.
    """
    close = buffer.close
    buffer.close = lambda: None
    yield buffer
    buffer.close = close
    buffer.seek(0)  # fake close


class FakeResponse(Response):
    def __init__(self, content: bytes = bytes(), status: int = 200):
        self._content = content
        self.status_code = status


class FakeWrite(Action[Name, int]):
    def __init__(self, file: FakeFile, to_write: bytes, delay: float):
        self.to_write = to_write
        self.delay = delay
        self.file = file

    def instruction(self) -> int:
        sleep(self.delay)
        result = self.file.write(self.to_write)
        return result
